package com.xydl.cac.iec;

import com.beanit.iec61850bean.*;
import com.xydl.cac.entity.IcdFile;
import com.xydl.cac.entity.IcdIed;
import com.xydl.cac.entity.constants.Constants;
import com.xydl.cac.exception.BusinessException;
import com.xydl.cac.model.StaticVariable;
import com.xydl.cac.repository.IcdFileRepository;
import com.xydl.cac.repository.IcdIedRepository;
import com.xydl.cac.socket.WebSocketServer;
import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Service;
import org.springframework.util.CollectionUtils;

import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;
import javax.annotation.Resource;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Optional;

@Service
@Slf4j
public class RealTimeDataService {
    @Resource
    IcdFileRepository fileRepository;
    @Resource
    IcdIedRepository iedRepository;
    @Resource
    WebSocketServer webSocketServer;

    public static boolean inDoing = false;
    HashMap<Integer, IecClient> clientMap = new HashMap<>();

    @PostConstruct
    private void init() {
        List<IcdIed> list = iedRepository.findByStart(Constants.TRUE);
        if (!CollectionUtils.isEmpty(list)) {
            for (IcdIed ied : list) {
                try {
                    this.startCollect(ied.getId());
                } catch (Exception ignore) {
                }
            }
        }
    }

    public void startCollect(Integer iedId) throws BusinessException {
        Optional<IcdIed> optional = iedRepository.findById(iedId);
        if (!optional.isPresent()) {
            throw new BusinessException("未找到该IED");
        }
        IcdIed ied = optional.get();
        Optional<IcdFile> optionalFile = fileRepository.findById(ied.getIcdFileId());
        if (!optionalFile.isPresent()) {
            throw new BusinessException("未找到该icd文件");
        }
        inDoing = true;
        IcdFile icdFile = optionalFile.get();
        try {
            IecClient iecClient = new IecClient();
            iecClient.connect(ied, icdFile.getXml());
            iecClient.enableReporting();
            iecClient.keep = true;
            iecClient.realTimeDataService = this;
            iecClient.webSocketServer = webSocketServer;
            log.info("61850订阅成功, ied=" + ied.getName() + ", ip=" + ied.getIp() + ", port=" + ied.getPort());
            ied.setStart(Constants.TRUE);
            iedRepository.save(ied);
            clientMap.put(ied.getId(), iecClient);
        } catch (Exception ex) {
            ied.setStart(Constants.FALSE);
            iedRepository.save(ied);
            String err = "61850订阅异常, ied=" + ied.getName() + ", ip=" + ied.getIp() + ", port=" + ied.getPort();
            log.error(err, ex);
            throw new BusinessException(err);
        } finally {
            inDoing = false;
        }
    }

    public void stopCollect(Integer iedId) {
        IecClient iecClient = clientMap.get(iedId);
        if (iecClient != null) {
            iecClient.disableReporting();
            iecClient.disconnect();
            clientMap.remove(iedId);
        }
        Optional<IcdIed> optional = iedRepository.findById(iedId);
        if (optional.isPresent()) {
            IcdIed ied = optional.get();
            ied.setStart(Constants.FALSE);
            iedRepository.save(ied);
            log.info("61850停止订阅, ied=" + ied.getName() + ", ip=" + ied.getIp());
        }
    }

    @PreDestroy
    private void stop() {
        log.info("关闭61850订阅服务.");
        Iterator<Integer> it = clientMap.keySet().iterator();
        while (it.hasNext()) {
            Integer key = it.next();
            IecClient iecClient = clientMap.get(key);
            iecClient.disableReporting();
            iecClient.disconnect();
            clientMap.remove(key);
        }
    }

    public static void processReport(Report report) {
        if (!CollectionUtils.isEmpty(report.getValues())) {
            for (FcModelNode node : report.getValues()) {
                try {
                    processNodeValue(node);
                } catch (Exception ignore) {
                }
            }
        }
    }

    private static void processNodeValue(FcModelNode node) {
        if (node instanceof FcDataObject) {
            FcDataObject fcnode = (FcDataObject) node;
            if (!CollectionUtils.isEmpty(fcnode.getChildren())) {
                for (ModelNode child : fcnode.getChildren()) {
                    if (child instanceof BasicDataAttribute) {
                        BasicDataAttribute bda = (BasicDataAttribute) child;
                        processBdaNodeValue(bda);
                    } else if (child instanceof ConstructedDataAttribute) {
                        ConstructedDataAttribute cda = (ConstructedDataAttribute) child;
                        if (!CollectionUtils.isEmpty(cda.getChildren())) {
                            for (ModelNode cchild : cda.getChildren()) {
                                if (cchild instanceof BasicDataAttribute) {
                                    BasicDataAttribute bda = (BasicDataAttribute) cchild;
                                    processBdaNodeValue(bda);
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    private static void processBdaNodeValue(BasicDataAttribute bda) {
        Iterator<Integer> it = StaticVariable.iecServerMap.keySet().iterator();
        while (it.hasNext()) {
            Integer key = it.next();
            IecServer iecServer = StaticVariable.iecServerMap.get(key);
            iecServer.updateBda(bda);
        }
        IEDCollectService.updateLastData(bda);
    }
}
